package de.hawhamburg.reachability.ui.frame;

import java.awt.Color;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.util.HashMap;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import de.hawhamburg.lifecycle.data.ConfigOrder;
import de.hawhamburg.lifecycle.data.JsonConfigOrder;
import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPPublisher;
import de.hawhamburg.reachability.ProjectInfo;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.enumeration.ReachabilityGroup;
import de.hawhamburg.reachability.enumeration.ReachabilityStatus;
import de.hawhamburg.reachability.json.JsonReachabilityInstruction;
import de.hawhamburg.reachability.tools.feeder.SensorDataFeeder;
import de.hawhamburg.reachability.tools.feeder.SensorDataFeederNotifiable;
import de.hawhamburg.reachability.ui.UI;
import de.hawhamburg.reachability.ui.frame.resource.ColorKeyUI;
import de.hawhamburg.reachability.ui.frame.resource.ColoredUIUpdater;
import de.hawhamburg.reachability.ui.frame.simulation.ReachabilitySimulationDetailsUI;
import de.hawhamburg.reachability.ui.frame.simulation.ReachabilityStartSimulationUI;
import de.hawhamburg.reachability.util.ConnectionUtil;
import de.hawhamburg.reachability.util.OutputUtil;

/**
 * The colored user interface to present the estimated reachability states to the inhabitant.<br><br>
 * 
 * This user interface provides full control over the reachability agent. It is able to let the agent start, stop or 
 * sleep. Furthermore it provides the actual estimated reachability states for the different groups encoded as color. It
 * has the ability to open an interface which describes the color to status mapping and to start and control
 * simulations.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class ColoredUI extends    UI
                       implements SensorDataFeederNotifiable {

    /**
     * Some unique serial version id.
     */
    private static final long serialVersionUID                  = 8178818625729508726L;

    /**
     * The current known reachability state.
     */
    private Map<ReachabilityGroup, ReachabilityStatus> status   = new HashMap<ReachabilityGroup, ReachabilityStatus>();
    
    /**
     * The {@link LPPublisher} to send the user feedback via the AMQ topic <i>REACHABILITY_AGENT_TOPIC</i>.
     */
    private LPPublisher sendUserFeedback                        = null;
    
    /**
     * The {@link LPPublisher} to send control messages like start, stop, sleep to the reachability agent via AMQ using
     * the topic <i>LP_CONTROL_TOPIC</i>. 
     */
    private LPPublisher controlTopic                            = null;
    
    /**
     * The {@link ColoredUIUpdater} which keeps the user interface up to date.
     */
    private ColoredUIUpdater updater                            = null;
    
    /**
     * The current running simulation tool as {@link SensorDataFeeder}.
     */
    private SensorDataFeeder simulationFeeder                   = null;
    
    /**
     * The menu item which opens the start simulation dialog.
     */
    private JMenuItem startSimulationMenuItem                   = null;
    
    /**
     * The menu item which opens the simulation control dialog.
     */
    private JMenuItem showSimulationMenuItem                    = null;

    /**
     * The different control {@link JPanel} for each {@link ReachabilityGroup} as {@link Map} for refreshing.
     */
    private Map<ReachabilityGroup, JPanel> colorPanels          =  new HashMap<ReachabilityGroup, JPanel>();
    
    /**
     * Creates a new colored reachability user interface.
     */
    public ColoredUI() {
        super("Colored Reachability Status");
        
        String feedbackTopicName = MainConfig.get("REACHABILITY_AGENT_TOPIC");
        String controlTopicName  = MainConfig.get("LP_CONTROL_TOPIC");
        
        sendUserFeedback = ConnectionUtil.getPublisher(feedbackTopicName, this);
        controlTopic     = ConnectionUtil.getPublisher(controlTopicName, this);
        
        updater = new ColoredUIUpdater(this);
        updater.start();
        
        build();
    }
    
    /**
     * This method adds a menu to the user interface.
     */
    protected final void addMenu() {
        JMenu control = new JMenu("Control");
        control.setMnemonic('C');
        
        JMenuItem start = new JMenuItem("Start");
        start.setMnemonic('S');
        start.setName("start");
        control.add(start);
        
        JMenuItem sleep = new JMenuItem("Sleep");
        sleep.setMnemonic('e');
        sleep.setName("sleep");
        control.add(sleep);
        
        JMenuItem stop = new JMenuItem("Stop");
        stop.setMnemonic('t');
        stop.setName("stop");
        control.add(stop);

        JMenu simulation = new JMenu("Simulation");
        simulation.setMnemonic('i');
        
        startSimulationMenuItem = new JMenuItem("Start simulation");
        startSimulationMenuItem.setMnemonic('s');
        startSimulationMenuItem.setName("startSimulation");
        simulation.add(startSimulationMenuItem);
        
        showSimulationMenuItem = new JMenuItem("Show simulation details");
        showSimulationMenuItem.setMnemonic('d');
        showSimulationMenuItem.setName("showSimulationDetails");
        showSimulationMenuItem.setEnabled(false);
        simulation.add(showSimulationMenuItem);
        
        JMenu help = new JMenu("Help");
        help.setMnemonic('H');
        
        JMenuItem legend = new JMenuItem("Color key");
        legend.setMnemonic('k');
        legend.setName("legend");
        help.add(legend);
        
        JMenuItem about = new JMenuItem("About Reachability Agent");
        about.setMnemonic('A');
        about.setName("about");
        help.add(about);
        
        JMenuBar bar = new JMenuBar();
        bar.add(control);
        bar.add(simulation);
        bar.add(help);
        setJMenuBar(bar);
        
        // Actions
        start.addActionListener(this);
        sleep.addActionListener(this);
        stop.addActionListener(this);
        startSimulationMenuItem.addActionListener(this);
        showSimulationMenuItem.addActionListener(this);
        legend.addActionListener(this);
        about.addActionListener(this);
    }
    
    /**
     * Returns a user interface grid.
     * 
     * @return              Returns a user interface grid.
     */
    private GridLayout getGrid() {
        return getGrid(1, ReachabilityGroup.values().length);
    }
    
    /**
     * Adds or updates the user interfaces status row with correct color to group mapping.
     */
    private void updateStatusRow() {
        for (ReachabilityGroup grp : ReachabilityGroup.values()) {
            // Get current state color
            ReachabilityStatus state = ReachabilityStatus.NOT_CALCULATED;
            if (status.containsKey(grp)) {
                state = status.get(grp);
            }
            Color stateColor = getStatusColor(state);
            
            // Set current state color
            if (colorPanels.containsKey(grp)) {
                JPanel groupPanel = colorPanels.get(grp);
                groupPanel.setBackground(stateColor);
                groupPanel.setToolTipText(state.toString());
            } else {
                JLabel label = new JLabel(grp.name());
                label.setOpaque(true);
                
                JPanel panel = new JPanel(new GridBagLayout());
                panel.setBackground(stateColor);
                panel.setToolTipText(state.toString());
                panel.add(label);
                panel.setBorder(BorderFactory.createLineBorder(Color.BLACK));
                
                addBody(panel);
                colorPanels.put(grp, panel);
            }
        }
    }
    
    /**
     * Reveals the color for a certain {@link ReachabilityStatus}.
     * 
     * @param theStatus     The {@link ReachabilityStatus} for which the color shall be determined.
     * @return              The {@link Color} corresponding to the {@link ReachabilityStatus}.
     */
    private Color getStatusColor(final ReachabilityStatus theStatus) {
        switch(theStatus) {
            case UNAVAILABLE:
                return Color.GRAY;
            case DO_NOT_DISTURB:
                return Color.RED;
            case BUSY:
                return Color.ORANGE;
            case AVAILABLE:
                return Color.GREEN;
            default:
                return Color.WHITE;
        }
    }
    
    /**
     * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
     * 
     * @param e             An {@link ActionEvent}.
     */
    public final void actionPerformed(final ActionEvent e) {
        String menuItemType = ((JMenuItem) e.getSource()).getName();
        if ("about".equals(menuItemType)) {
            infoDialog("Version: " + MainConfig.get("REACHABILITY_UI_VERSION")
                     + "\nAuthor: " + ProjectInfo.REACHABILITY_AGENT_AUTHOR);
        }
        if ("legend".equals(menuItemType)) {
            displayColorKey();
        }
        if ("start".equals(menuItemType) || "sleep".equals(menuItemType) || "stop".equals(menuItemType)) {
            sendControl(ConfigOrder.valueOf(menuItemType));
        }
        if ("startSimulation".equals(menuItemType)) {
            startSimulationDialog();
        }
        if ("showSimulationDetails".equals(menuItemType)) {
            showSimulationDetailsDialog();
        }
    }

    /**
     * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
     * 
     * @param e             A {@link MouseEvent}.
     */
    public final void mousePressed(final MouseEvent e) {
        JComponent src = (JComponent) e.getSource();
        String     key = getConfigValue("FIELD_LABEL_PREFIX_JLIST");
        
        if (key.equals(src.getName().substring(0, key.length()))) {
            @SuppressWarnings("unchecked")
            JList<ReachabilityStatus> list      = (JList<ReachabilityStatus>) e.getSource();
            ReachabilityGroup tempGroup         = ReachabilityGroup.valueOf(list.getName().substring(key.length()));
            ReachabilityStatus tempStatus       = (ReachabilityStatus) list.getSelectedValue();
            JsonReachabilityInstruction inst    = getInstructionMsg(tempGroup, tempStatus);
            ConnectionUtil.setMessageOnTopic(sendUserFeedback, inst);
        }
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     */
    public final void shutDown() {
        // Shut down other threads
        if (updater != null) {
            updater.shutDown();
        }
        
        // Shut down own AMQ publisher/subscriber
        if (sendUserFeedback != null) {
            sendUserFeedback.disconnect();
        }
        if (controlTopic != null) {
            controlTopic.disconnect();
        }
    }
    
    /**
     * Open a new dialog to start a simulation.
     */
    private void startSimulationDialog() {
        new ReachabilityStartSimulationUI(this);
    }
    
    /**
     * Opens a new dialog to show the current simulation state.
     */
    private void showSimulationDetailsDialog() {
        new ReachabilitySimulationDetailsUI(this, simulationFeeder);
    }
    
    /**
     * Opens a new dialog which shows the mapping of {@link Color} to {@link ReachabilityStatus}.
     */
    private void displayColorKey() {
        Map<ReachabilityStatus, Color> map = new HashMap<ReachabilityStatus, Color>();
        for (ReachabilityStatus state : ReachabilityStatus.values()) {
            map.put(state, getStatusColor(state));
        }
        new ColorKeyUI(this, map);
    }
    
    /**
     * Initial build of this user interface. The rendering won't be influenced by the <i>refresh</i> parameter.
     * 
     * @param refresh       Indicates if this method was called with the intention to re-render the interface. 
     */
    public final void refresh(final boolean refresh) {
        refresh(null);
    }
    
    /**
     * Refreshes the cached {@link ReachabilityStatus} and the user interface.
     * 
     * @param newStatus     A hash {@link Map} containing the current {@link ReachabilityStatus}.
     */
    public final void refresh(final Map<ReachabilityGroup, ReachabilityStatus> newStatus) {
        if (newStatus != null) {
            this.status = newStatus;
        } else {
            getBody().setLayout(getGrid());
            getContentPane().add(getBody());
            getBody().removeAll();
        }
        updateStatusRow();
        getBody().revalidate();
    }
    
    /**
     * Calls {@link #getInstructionMsg(ReachabilityGroup, ReachabilityStatus, String)} with a blank {@link String} as 
     * remark and returns its result.
     * 
     * @param theGroup      A {@link ReachabilityGroup}.
     * @param theStatus     A {@link ReachabilityStatus}
     * @return              A {@link JsonReachabilityInstruction} message.
     */
    public final JsonReachabilityInstruction getInstructionMsg(final ReachabilityGroup theGroup,
                                                               final ReachabilityStatus theStatus) {
        return getInstructionMsg(theGroup, theStatus, "");
    }
    
    /**
     * Generates a {@link JsonReachabilityInstruction} messages and returns it.
     * 
     * @param theGroup      A {@link ReachabilityGroup}
     * @param theStatus     A {@link ReachabilityStatus}
     * @param theRemark     A remark as {@link String}.
     * @return              A {@link JsonReachabilityInstruction} message.
     */
    public final JsonReachabilityInstruction getInstructionMsg(final ReachabilityGroup theGroup,
                                                               final ReachabilityStatus theStatus,
                                                               final String theRemark) {
        JsonReachabilityInstruction inst = new JsonReachabilityInstruction();
        inst.Id         = getId();
        inst.Version    = MainConfig.get("REACHABILITY_UI_VERSION");
        inst.Group      = theGroup;
        inst.Status     = theStatus;
        inst.Remark     = theRemark;
        return inst;
    }
    
    /**
     * Sends a {@link ConfigOrder} in a {@link JsonConfigOrder} messages via AMQ to the reachability agent.
     * 
     * @param x             The {@link ConfigOrder} which shall be send.
     */
    public final void sendControl(final ConfigOrder x) {
        JsonConfigOrder order = new JsonConfigOrder();
        order.Version = MainConfig.get("REACHABILITY_UI_VERSION");
        order.Id      = ProjectInfo.REACHABILITY_AGENT_NAME;
        order.order   = x;
        ConnectionUtil.setMessageOnTopic(controlTopic, order);
    }
    
    /**
     * Registers a new simulation feeder. Disables the start simulation and enables the simulation controlling menu 
     * option.
     * 
     * @param feeder        The new simulation feeder as {@link SensorDataFeeder} which shall be registered.
     */
    public final void registerSimulationFeeder(final SensorDataFeeder feeder) {
        this.simulationFeeder = feeder;
        this.startSimulationMenuItem.setEnabled(false);
        this.showSimulationMenuItem.setEnabled(true);
    }
    
    /**
     * Unregisters a simulation feeder. Disables the simulation controlling and enables the start start simulation menu
     * option.
     * 
     * @param feeder        The simulation feeder as {@link SensorDataFeeder} which shall be unregistered.
     */
    public final void unregisterSimulationFeeder(final SensorDataFeeder feeder) {
        if (this.simulationFeeder.equals(feeder)) {
            this.simulationFeeder = null;
            this.startSimulationMenuItem.setEnabled(true);
            this.showSimulationMenuItem.setEnabled(false);
        }
    }
    
    /**
     * Starts a new {@link ColoredUI} in the systems look and feel.
     * 
     * @param args              The system supplied start arguments as array of {@link String}.
     */
    public static void main(final String[] args) {
        // Remove all existing loggers
        org.apache.log4j.Logger.getRootLogger().removeAllAppenders();
        
        // Try to adapt the systems look and feel
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (ClassNotFoundException e) {
            OutputUtil.error("[ClassNotFoundException] " + e.getMessage());
        } catch (InstantiationException e) {
            OutputUtil.error("[InstantiationException] " + e.getMessage());
        } catch (IllegalAccessException e) {
            OutputUtil.error("[IllegalAccessException] " + e.getMessage());
        } catch (UnsupportedLookAndFeelException e) {
            OutputUtil.error("[UnsupportedLookAndFeelException] " + e.getMessage());
        }
        
        // Start a new color user interface
        new ColoredUI();
    }
}
