package de.hawhamburg.reachability.ui.frame.resource;

import java.awt.Color;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.util.HashMap;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import de.hawhamburg.reachability.ProjectInfo;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.enumeration.ReachabilityStatus;
import de.hawhamburg.reachability.ui.UI;
import de.hawhamburg.reachability.util.OutputUtil;

/**
 * User interface to explain which color is related to which {@link ReachabilityStatus}.<br><br>
 * 
 * This user interface will show a single columned table with as many rows as the amount of distinct 
 * {@link ReachabilityStatus}. Each row will contain the name of the {@link ReachabilityStatus} and be colored in the
 * color of this {@link ReachabilityStatus}.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class ColorKeyUI extends UI {
    
    /**
     * Some unique serial version id.
     */
    private static final long serialVersionUID          = 680298812488691640L;
    
    /**
     * The color-key map, holding the color to each {@link ReachabilityStatus}.
     */
    private Map<ReachabilityStatus, Color> colorKeyMap  = new HashMap<ReachabilityStatus, Color>();
    
    /**
     * Creates and builds a new {@link ColorKeyUI}.
     * 
     * @param sourceWindow  The calling panel which will be notified, when the simulation was started as Object 
     *                      implementing {@link JFrame}.
     * @param colorMapping  The mapping of {@link ReachabilityStatus} to {@link Color} will be supplied as initial 
     *                      parameter.
     */
    public ColorKeyUI(final JFrame sourceWindow,
                      final Map<ReachabilityStatus, Color> colorMapping) {
        super("Color Key", sourceWindow);
        this.colorKeyMap = colorMapping;
        build();
    }
    
    /**
     * Creates and builds a new {@link ColorKeyUI} without source window by using {@link #ColorKeyUI(JFrame, Map)}.
     * 
     * @param colorMapping  The mapping of {@link ReachabilityStatus} to {@link Color} will be supplied as initial 
     *                      parameter.
     */
    public ColorKeyUI(final Map<ReachabilityStatus, Color> colorMapping) {
        this(null, colorMapping);
    }
    
    /**
     * This method adds a menu to the user interface.
     */
    protected final void addMenu() {
        if (this.isStandalone()) {
            JMenu help = new JMenu("Help");
            help.setMnemonic('H');
            
            JMenuItem about = new JMenuItem("About Reachability Agent");
            about.setMnemonic('A');
            about.setName("about");
            help.add(about);
            
            JMenuBar bar = new JMenuBar();
            bar.add(help);
            setJMenuBar(bar);
            
            // Actions
            about.addActionListener(this);
        }
    }
    
    /**
     * Returns a user interface grid.
     * 
     * @param rowsIn        Amount of rows to be displayed
     * @return              Returns a user interface grid.
     */
    private GridLayout getGrid(final int rowsIn) {
        return getGrid(rowsIn, 1);
    }
    
    /**
     * Adds a new row to the users interface, containing the a {@link ReachabilityStatus} and its color.
     * 
     * @param state         The {@link ReachabilityStatus} for which the color shall added as row to the user interface.
     */
    private void addInfoRow(final ReachabilityStatus state) {
        if (colorKeyMap.containsKey(state)) {
            JLabel label = new JLabel(state.toString());
            label.setOpaque(true);
            
            JPanel panel = new JPanel(new GridBagLayout());
            panel.setBackground(colorKeyMap.get(state));
            panel.setToolTipText(state.toString());
            panel.add(label);
            panel.setBorder(BorderFactory.createLineBorder(Color.BLACK));
            
            addBody(panel);
        }
    }
        
    /**
     * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
     * 
     * @param e             An {@link ActionEvent}.
     */
    public final void actionPerformed(final ActionEvent e) {
        // Menu Items
        if (JMenuItem.class.isAssignableFrom(e.getSource().getClass())) {
            String menuItemType = ((JMenuItem) e.getSource()).getName();
            if ("about".equals(menuItemType)) {
                infoDialog("Version: " + MainConfig.get("REACHABILITY_UI_VERSION")
                        + "\nAuthor: " + ProjectInfo.REACHABILITY_AGENT_AUTHOR);
            }
        }
    }
    
    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     */
    public final void shutDown() {
        // Shut down everything...
    }
    
    /**
     * Builds the user interface. The rendering will not be influenced by the <i>refresh</i> parameter.
     * 
     * @param refresh       Indicates if this method was called intentionally to initialize the user infertace of not. 
     */
    public final void refresh(final boolean refresh) {
        int rows = colorKeyMap.size();
                
        // Adapt window size
        int newHeight = getConfigValueAsInteger("DEFAULT_WINDOW_MENU_HEIGHT")
                        + (getConfigValueAsInteger("DEFAULT_ROW_HEIGHT") * rows);
        setWindowHeight(newHeight);
        fixBounds(refresh);
        
        // Define layout
        if (!refresh) {
            getBody().setLayout(getGrid(rows));
            getContentPane().add(getBody());
            getBody().removeAll();
        }
        
        // Set story rows
        for (ReachabilityStatus state : colorKeyMap.keySet()) {
            addInfoRow(state);
        }
            
        // Reload content
        fixBounds(refresh);
        getBody().revalidate();
    }
    
    /**
     * Starts a new {@link ColorKeyUI} in the systems look and feel for testing reasons. This user interfaces wasn't
     * primarily defined as stand-alone interface. 
     * 
     * @param args              The systems supplied start arguments as array of {@link String}.
     */
    public static void main(final String[] args) {
        // Remove all existing loggers
        org.apache.log4j.Logger.getRootLogger().removeAllAppenders();
        
        // Try to show the user interface in systems look and feel
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (ClassNotFoundException e) {
            OutputUtil.error("[ClassNotFoundException] " + e.getMessage());
        } catch (InstantiationException e) {
            OutputUtil.error("[InstantiationException] " + e.getMessage());
        } catch (IllegalAccessException e) {
            OutputUtil.error("[IllegalAccessException] " + e.getMessage());
        } catch (UnsupportedLookAndFeelException e) {
            OutputUtil.error("[UnsupportedLookAndFeelException] " + e.getMessage());
        }

        // Start and build the user interfaces with a dummy status to color mapping
        Map<ReachabilityStatus, Color> map = new HashMap<ReachabilityStatus, Color>();
        map.put(ReachabilityStatus.AVAILABLE,       Color.GREEN);
        map.put(ReachabilityStatus.BUSY,            Color.ORANGE);
        map.put(ReachabilityStatus.DO_NOT_DISTURB,  Color.RED);
        map.put(ReachabilityStatus.UNAVAILABLE,     Color.GRAY);
        map.put(ReachabilityStatus.NOT_CALCULATED,  Color.WHITE);
        new ColorKeyUI(map);
    }
}
