package de.hawhamburg.reachability.ui.frame.resource;

import java.util.Map;
import java.util.HashMap;

import com.google.gson.Gson;

import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPSubscriber;
import de.hawhamburg.reachability.StopableObject;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.enumeration.ReachabilityGroup;
import de.hawhamburg.reachability.enumeration.ReachabilityStatus;
import de.hawhamburg.reachability.json.JsonReachabilityInformation;
import de.hawhamburg.reachability.ui.frame.ColoredUI;
import de.hawhamburg.reachability.util.ConnectionUtil;

/**
 * Updater thread for the colored UI.<br><br>
 * 
 * This thread shall keep the {@link ColoredUI} up to date.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public class ColoredUIUpdater extends Thread implements StopableObject {
    
    /**
     * The {@link ColoredUI} object, which shall be updated.
     */
    private ColoredUI ui                                                = null;

    /**
     * Determines if the method {@link #shutDown()} has been called.
     */
    private boolean shutDown                                            = false;
    
    /**
     * An {@link LPSubscriber}, which listens on the reachability status topic (<i>REACHABILITY_AGENT_RESULT_TOPIC</i>).
     */
    private LPSubscriber infoTopic                                      = null;
    
    /**
     * An hash map holding the last known states of reachability for each {@link ReachabilityGroup} item.
     */
    private Map<ReachabilityGroup, ReachabilityStatus> internalStatus   =
            new HashMap<ReachabilityGroup, ReachabilityStatus>();
    
    /**
     * Generates a new {@link ColoredUIUpdater} thread to keep the {@link ColoredUI} <i>ui</i> up to date.
     * 
     * @param sourceUI      A {@link ColoredUI}, which shall be updated.
     */
    public ColoredUIUpdater(final ColoredUI sourceUI) {
        this.ui = sourceUI;
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     */
    public final void shutDown() {
        this.shutDown = true;
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#isActive()
     * 
     * @return              <code>true</code> if {@link #shutDown()} wasn't called, otherwise <code>false</code>.
     */
    public final boolean isActive() {
        return !this.shutDown;
    }

    /**
     * Connects to the reachability information topic using {@link #init()} and frequently tries to get the newest 
     * reachability states from this topic. If a new state arrives, the {@link ColoredUI} {@link #ui} will be updated 
     * using {@link ColoredUI#refresh(Map)}.
     */
    public final void run() {
        init();
        
        while (isActive()) {
            
            String msg = ConnectionUtil.getMessageOnTopic(infoTopic);
            if (msg != null) {
                Gson g = new Gson();
                JsonReachabilityInformation info = g.fromJson(msg, JsonReachabilityInformation.class);
                for (Map.Entry<ReachabilityGroup, ReachabilityStatus> state : info.currentStatus.entrySet()) {
                    if (internalStatus.containsKey(state.getKey())) {
                        internalStatus.put(state.getKey(), state.getValue());
                    }
                }
                ui.refresh(internalStatus);
            }
            
        }
        
        end();
    }
    
    /**
     * Opens a connection the the reachability information topic (<i>REACHABILITY_AGENT_RESULT_TOPIC</i>) and 
     * initializes the intern state hash map {@link #internalStatus} with {@link ReachabilityStatus#NOT_CALCULATED}.
     */
    public final void init() {
        // Fetch topic names
        String infoTopicName = MainConfig.get("REACHABILITY_AGENT_RESULT_TOPIC");
        
        // Open topics
        infoTopic = ConnectionUtil.getSubscriber(infoTopicName, this);
        
        // Preset status
        for (ReachabilityGroup x : ReachabilityGroup.values()) {
            internalStatus.put(x, ReachabilityStatus.NOT_CALCULATED);
        }
    }
    
    /**
     * Disconnects the {@link #infoTopic}, if it isn't <code>null</code>.
     */
    public final void end() {
        if (infoTopic != null) {
            infoTopic.disconnect();
        }
    }
}
